// Haza's CVar Extras Plugin

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <interface.h>
#include "icvar.h"
#include "convar.h"
#include "iclient.h"
#include "SigScan.h"
#include "inetchannel.h"
#include "inetchannelinfo.h"
#include "bitbuf.h"

#include <string>

#include "GMLuaModule.h"

GMOD_MODULE(Load, Unload)
ILuaInterface* gLua;

ICvar **cvar2;

/* Gate Keeper Code */
class CSomeEngineClass; // 'class CSomeEngineClass : public IClient'
CSomeEngineClass*** ppClientList;
int* pClientListSize;
/* End Gate Keeper Code: Thanks alot :)*/

#define FCVAR_DEVELOPMENTONLY	(1<<1)

LUA_FUNCTION(SetFlags)
{
	gLua->CheckType(1, GLua::TYPE_STRING);
	gLua->CheckType(2, GLua::TYPE_NUMBER);

	if(!cvar2)
	{
		gLua->Error("cvar2 interface not loaded.");
		return 0;
	}

	ConCommandBase *var = (*cvar2)->FindCommandBase(gLua->GetString(1));

	if(!var)
	{
		gLua->Error("cvar/ccmd not found.");
		return 0;
	}

	// m_nFlags needs to be made public in convar.h, this will only affect this dll, the class struct will stay the same but the compiler won't complain.
	int *flags = &var->m_nFlags;
	*flags = gLua->GetInteger(2);

	return 0;
}

LUA_FUNCTION(GetFlags)
{
	gLua->CheckType(1, GLua::TYPE_STRING);

	if(!cvar2)
	{
		gLua->Error("cvar2 interface not loaded.");
		return 0;
	}

	ConCommandBase *var = (*cvar2)->FindCommandBase(gLua->GetString(1));

	if(!var)
	{
		gLua->Error("cvar/ccmd not found.");
		return 0;
	}

	// m_nFlags needs to be made public in convar.h, this will only affect this dll, the class struct will stay the same but the compiler won't complain.
	gLua->PushDouble((int)var->m_nFlags);

	return 1;
}

LUA_FUNCTION(ToggleFlag)  
{  
	gLua->CheckType(1, GLua::TYPE_STRING);  
	gLua->CheckType(2, GLua::TYPE_NUMBER);  

	if(!cvar2)  
	{  
		gLua->Error("cvar2 interface not loaded.");  
		return 0;  
	}  

	ConCommandBase *var = (*cvar2)->FindCommandBase(gLua->GetString(1));  

	if(!var)  
	{  
		gLua->Error("cvar/ccmd not found.");  
		return 0;  
	}  

	// m_nFlags needs to be made public in convar.h, this will only affect this dll, the class struct will stay the same but the compiler won't complain.  
	int *flags = &var->m_nFlags;  

	// Thanks to hfs for this section :)
	*flags ^= gLua->GetInteger(2);  

	gLua->Push((bool)(*flags & gLua->GetInteger(2)));

	return 1;  
}  

LUA_FUNCTION(GetAll)
{
	if(!cvar2)
	{
		gLua->Error("cvar2 interface not loaded.");
		return 0;
	}

	ConCommandBase *var = (*cvar2)->GetCommands();

	if(!var)
	{
		gLua->Error("cvar2 GetAll error");
		return 0;
	}

	ILuaObject* cvarTbl = gLua->GetNewTable();

	int incr = 1;
	while(var)
	{
		cvarTbl->SetMember(incr, var->GetName());
		var = var->GetNext();
		incr++;
	}

	gLua->Push(cvarTbl);

	cvarTbl->UnReference();

	return 1;
}

LUA_FUNCTION(SetHelpText)
{
	gLua->CheckType(1, GLua::TYPE_STRING);
	gLua->CheckType(2, GLua::TYPE_STRING);

	if(!cvar2)
	{
		gLua->Error("cvar2 interface not loaded.");
		return 0;
	}

	ConCommandBase *var = (*cvar2)->FindCommandBase(gLua->GetString(1));

	if(!var)
	{
		gLua->Error("cvar/ccmd not found.");
		return 0;
	}

	// This is really bad, call this too much, then you will have a memleak. Will fix one day.
	std::string tmp = gLua->GetString(2);
	const char *mem = (const char *)malloc(tmp.size() + 1);
	memcpy((void *)mem, tmp.c_str(), tmp.size());
	((char *)mem)[tmp.size()] = 0x00;

	// m_pszHelpString needs to be made public in convar.h, this will only affect this dll, the class struct will stay the same but the compiler won't complain.
	var->m_pszHelpString = &mem[0];

	return 0;
}

LUA_FUNCTION(GetValue)
{
	gLua->CheckType(1, GLua::TYPE_STRING);

	if(!cvar2)
	{
		gLua->Error("cvar2 interface not loaded.");
		return 0;
	}

	ConVar *var = (*cvar2)->FindVar(gLua->GetString(1));

	if(!var)
	{
		gLua->Error("cvar not found.");
		return 0;
	}

	gLua->Push(var->GetString());

	return 1;
}

LUA_FUNCTION(IsFlagSet)
{
	gLua->CheckType(1, GLua::TYPE_STRING);
	gLua->CheckType(2, GLua::TYPE_NUMBER);

	if(!cvar2)
	{
		gLua->Error("cvar2 interface not loaded.");
		return 0;
	}

	ConVar *var = (*cvar2)->FindVar(gLua->GetString(1));

	if(!var)
	{
		gLua->Error("cvar not found.");
		return 0;
	}

	gLua->Push((bool)var->IsFlagSet(gLua->GetInteger(2)));

	return 1;
}

LUA_FUNCTION(SetValue)
{
	gLua->CheckType(1, GLua::TYPE_STRING);

	int dtType = gLua->GetType(2);

	if(dtType == GLua::TYPE_NIL || dtType == GLua::TYPE_INVALID)
	{
		gLua->Error("arg[2] invalid");
		return 0;
	}

	if(!cvar2)
	{
		gLua->Error("cvar2 interface not loaded.");
		return 0;
	}

	ConVar *var = (*cvar2)->FindVar(gLua->GetString(1));

	if(!var)
	{
		gLua->Error("cvar not found.");
		return 0;
	}

	switch(dtType)
	{
	case GLua::TYPE_NUMBER:
		var->SetValue((float)gLua->GetDouble(2));
		break;
	case GLua::TYPE_BOOL:
		var->SetValue(gLua->GetBool(2) ? 1 : 0);
		break;
	case GLua::TYPE_STRING:
		var->SetValue(gLua->GetString(2));
		break;
	default:
		gLua->Error("arg[2] invalid");
		return 0;
		break;
	}

	return 0;
}

LUA_FUNCTION(ResetValue)
{
	gLua->CheckType(1, GLua::TYPE_STRING);

	if(!cvar2)
	{
		gLua->Error("cvar2 interface not loaded.");
		return 0;
	}

	ConVar *var = (*cvar2)->FindVar(gLua->GetString(1));

	if(!var)
	{
		gLua->Error("cvar not found.");
		return 0;
	}

	var->Revert();

	return 0;
}

LUA_FUNCTION(ExecuteCommandOnClient)
{
	gLua->CheckType(1, GLua::TYPE_NUMBER);
	gLua->CheckType(2, GLua::TYPE_STRING);

	if(!pClientListSize)
		return 0;

	int clientId = gLua->GetInteger(1);
	
	for(int i = 0; i < (*pClientListSize); i++)
	{
		IClient *c = (IClient *)((size_t)(*ppClientList)[i] + 4); // Gate Keeper

		if(c->GetUserID() - 1 == clientId)
		{
			c->ExecuteStringCommand(gLua->GetString(2));
			break;
		}
	}

	return 0;
}

LUA_FUNCTION(GetUserSetting)
{
	gLua->CheckType(1, GLua::TYPE_NUMBER);
	gLua->CheckType(2, GLua::TYPE_STRING);

	if(!pClientListSize)
		return 0;

	int clientId = gLua->GetInteger(1);

	for(int i = 0; i < (*pClientListSize); i++)
	{
		IClient *c = (IClient *)((size_t)(*ppClientList)[i] + 4); // Gate Keeper

		if(c->GetUserID() - 1 == clientId)
		{
			gLua->Push(c->GetUserSetting(gLua->GetString(2)));
			return 1;
		}
	}

	return 0;
}

LUA_FUNCTION(ReplicateData)
{
	gLua->CheckType(1, GLua::TYPE_NUMBER);
	gLua->CheckType(2, GLua::TYPE_STRING);
	gLua->CheckType(3, GLua::TYPE_STRING);

	if(!pClientListSize)
		return 0;
	
	int clientId = gLua->GetInteger(1);

	for(int i = 0; i < (*pClientListSize); i++)
	{
		IClient *c = (IClient *)((size_t)(*ppClientList)[i] + 4); // Gate Keeper

		if(c->GetUserID() - 1 == clientId)
		{
			INetChannel *netChan = c->GetNetChannel();
			char pckBuf[256];
			bf_write pck(pckBuf, sizeof(pckBuf));

			pck.WriteUBitLong(5/* NET_SetConVar */, 5);
			pck.WriteByte(0x01);
			pck.WriteString(gLua->GetString(2)/* CVar name */);
			pck.WriteString(gLua->GetString(3)/* Value as string */);

			netChan->SendData(pck);
			
			break;
		}
	}

	return 0;
}

int Load(lua_State* L)
{
	gLua = Lua();

	Sys_LoadModule("server.dll");

	// cvar is the only(usefull) DLL export made in server.dll client.dll
	// this is better than getting a new interface to this etc.
	cvar2 = (ICvar **)GetProcAddress(GetModuleHandleA("server.dll"), "cvar");

	/* Gate Keeper Code */
	CSigScan::sigscan_dllfunc = Sys_GetFactory("engine.dll");
	CSigScan::GetDllMemInfo();

	CSigScan sigClientList;
	sigClientList.Init((unsigned char *) "\x7c\x86\xee\x0c\x8b\x04\xba\x85\xc0\x74\x05\x8d\x70\x04\xEB\x02\x33\xf6\x8b\x06\x8B\x50\x78\x8B\xCE",
		"????xxxxxxxxxxxxxxxxxxxxx", 25);

	if ( !sigClientList.is_set )
		gLua->Error("ClientList signature scan failed!");

	ppClientList = *(CSomeEngineClass ****) sigClientList.sig_addr;
	pClientListSize = (int *) ((int) ppClientList + 0xC);
	/* End Gate Keeper Code: Thanks alot :)*/

	ILuaObject* cvarTbl = gLua->GetNewTable();
	cvarTbl->SetMember("SetFlags", SetFlags);
	cvarTbl->SetMember("GetFlags", GetFlags);
	cvarTbl->SetMember("SetHelpText", SetHelpText);
	cvarTbl->SetMember("GetAll", GetAll);
	cvarTbl->SetMember("ToggleFlag", ToggleFlag);
	cvarTbl->SetMember("GetValue", GetValue);
	cvarTbl->SetMember("SetValue", SetValue);
	cvarTbl->SetMember("IsFlagSet", IsFlagSet);
	cvarTbl->SetMember("ResetValue", ResetValue);
	cvarTbl->SetMember("ExecuteCommandOnClient", ExecuteCommandOnClient);
	cvarTbl->SetMember("GetUserSetting", GetUserSetting);
	cvarTbl->SetMember("ReplicateData", ReplicateData);
	gLua->SetGlobal("cvar2", cvarTbl);
	cvarTbl->UnReference();

	gLua->SetGlobal("FCVAR_DEVELOPMENTONLY", (float)FCVAR_DEVELOPMENTONLY);

	if(!cvar2)
		gLua->Error("cvar2 interface not loaded.");

	return 0;
}

int Unload(lua_State* L)
{
	return 0;
}